#ifndef X11_H
#define X11_H

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>

#include <string>


// Relative coordinates between (0; 0) for top-left and (max-1; max-1)
// for bottom-right.
class RCoord {
	int x_, y_;
public:
	enum { max = 1024 };

	RCoord() {}
	RCoord(int x, int y) : x_(x), y_(y) {}
	int X() const { return x_; }
	int Y() const { return y_; }
	bool XOut() const { return x_ < 0 || x_ >= max; }
	bool YOut() const { return y_ < 0 || y_ >= max; }
	bool Out() const { return XOut() || YOut(); }
	RCoord XMirror() const { return RCoord(max - x_ - 1, y_); }
	RCoord & operator+=(RCoord rhs)
		{ x_ += rhs.x_; y_ += rhs.y_; return *this; }
	RCoord & operator-=(RCoord rhs)
		{ x_ -= rhs.x_; y_ -= rhs.y_; return *this; }
};

inline RCoord operator+(RCoord lhs, RCoord rhs) { return lhs += rhs; }
inline RCoord operator-(RCoord lhs, RCoord rhs) { return lhs -= rhs; }


// Absolute coordinates
class Coord {
	XPoint coord_;
public:
	Coord() {}
	Coord(short x, short y) { coord_.x = x;	coord_.y = y; }
	Coord(RCoord);
	RCoord ToRCoord() const;
	operator       XPoint&()       { return coord_; }
	operator const XPoint&() const { return coord_; }
	int X() const { return coord_.x; }
	int Y() const { return coord_.y; }
	void SetX(short x) { coord_.x = x; }
	void SetY(short y) { coord_.y = y; }
	void AddX(int x) { coord_.x += x; }
	void AddY(int y) { coord_.y += y; }
	bool XOut() const;
	bool YOut() const;
	bool Out() const { return XOut() || YOut(); }
	double Length() const;
	Coord & operator+=(Coord rhs) {
		coord_.x += rhs.coord_.x;
		coord_.y += rhs.coord_.y;
		return *this;
	}
	Coord & operator-=(Coord rhs) {
		coord_.x -= rhs.coord_.x;
		coord_.y -= rhs.coord_.y;
		return *this;
	}
	template<class T> Coord & operator*=(T rhs) {
		coord_.x = static_cast<short>(coord_.x * rhs);
		coord_.y = static_cast<short>(coord_.y * rhs);
		return *this;
	}
	template<class T> Coord & operator/=(T rhs) {
		coord_.x = static_cast<short>(coord_.x / rhs);
		coord_.y = static_cast<short>(coord_.y / rhs);
		return *this;
	}
};

inline Coord operator+(Coord lhs, Coord rhs) { return lhs += rhs; }
inline Coord operator-(Coord lhs, Coord rhs) { return lhs -= rhs; }
template<class T> inline Coord operator*(T lhs, Coord rhs)
	{ return rhs *= lhs; }
template<class T> inline Coord operator/(Coord lhs, T rhs)
	{ return lhs /= rhs; }



// Implementations of conversions from pixels distances to relative distances.
inline int Rel(int dist, int scale)
	{ return (dist * RCoord::max + scale / 2) / scale; }
inline int UnRel(int r, int scale)
	{ return (r * scale + RCoord::max / 2) / RCoord::max; }



// The graphic interface is the window. It is possible to draw inside
// and to get events.
class X11 {
	static X11 * singleton_;
	static unsigned default_window_width_;
	static unsigned default_window_height_;

	X11();
	~X11();

	Display * display_;
	int screen_;
	Colormap colormap_;
	GC gc_;
	Window window_;
	// cached values for relative coordinates mapping
	unsigned window_width_, window_height_;
public:
	typedef unsigned long Color;

	static X11 & Inst();
	static void CreateInstance();
	static void DestroyInstance();
	static void SetDefaultWindowWidth(int);
	static void SetDefaultWindowHeight(int);

	void SetWindowTitle(const std::string & title) const;
	Status GetWindowAttributes(XWindowAttributes * = 0);
	int WindowWidth() const;
	int WindowHeight() const;

	Color GetBlack() const;
	Color GetWhite() const;
	void SetForeground(Color color) const;
	Status AllocColor(XColor *) const;
	void FreeColor(Color) const;

	void DrawPoint(int x, int y) const;
	void DrawPoint(Coord) const;
	void DrawPoints(const XPoint * points, int npoints) const;
	void DrawPoints(const Coord * points, int npoints) const;
	void SetClipOrigin(int x, int y) const;
	void SetClipOrigin(Coord) const;
	void SetClipMask(Pixmap) const;
	void ClearArea(int x, int y, unsigned width, unsigned height,
                       Bool exposures = False) const;
	void ClearArea(Coord c, unsigned width, unsigned height,
                       Bool exposures = False) const;
	void CopyAreaToWindow(Drawable src, int src_x, int src_y,
	                      unsigned width, unsigned height,
		              int dest_x, int dest_y) const;
	void CopyAreaToWindow(Drawable src, int src_x, int src_y,
	                      unsigned width, unsigned height,
			      Coord dest_c) const;
	void NextEvent(XEvent *) const;
	int Pending() const;
	void Sync(Bool discard) const;

	int CreatePixmapFromData(const char ** xpm, Pixmap * pixmap,
	                         Pixmap * pixmask, XpmAttributes *) const;
	void FreePixmap(Pixmap) const;
	XFontStruct * LoadQueryFont(const std::string & font_name) const;
	void FreeFont(XFontStruct * font) const;
	void SetFont(Font font) const;
	void DrawString(Coord c, const std::string & str) const;
};


inline X11 & X11::Inst()
	{ return *singleton_; }
inline void X11::CreateInstance()
	{ if (!singleton_) singleton_ = new X11; }
inline void X11::DestroyInstance()
	{ delete singleton_; singleton_ = 0; }
inline void X11::SetDefaultWindowWidth(int width)
	{ default_window_width_ = width; }
inline void X11::SetDefaultWindowHeight(int height)
	{ default_window_height_ = height; }
inline int X11::WindowWidth() const
	{ return window_width_; }
inline int X11::WindowHeight() const
	{ return window_height_; }
inline X11::Color X11::GetBlack() const
	{ return BlackPixel(display_, screen_); }
inline X11::Color X11::GetWhite() const
	{ return WhitePixel(display_, screen_); }
inline void X11::SetForeground(Color color) const
	{ XSetForeground(display_, gc_, color); }
inline Status X11::AllocColor(XColor * color) const
	{ return XAllocColor(display_, colormap_, color); }
//inline void X11::FreeColor(Color color) const
//	{ XFreeColors(display_, colormap_, &color, 1, 0); }
inline void X11::DrawPoint(int x, int y) const
	{ XDrawPoint(display_, window_, gc_, x, y); }
inline void X11::DrawPoint(Coord c) const
	{ DrawPoint(c.X(), c.Y()); }
inline void X11::DrawPoints(const XPoint * points, int npoints) const
{
	XDrawPoints(display_, window_, gc_, const_cast<XPoint*>(points),
	            npoints, CoordModeOrigin);
}
inline void X11::DrawPoints(const Coord * points, int npoints) const
	{ DrawPoints(&(points->operator const XPoint&()), npoints); }
inline void X11::SetClipOrigin(int x, int y) const
	{ XSetClipOrigin(display_, gc_, x, y); }
inline void X11::SetClipOrigin(Coord c) const
	{ XSetClipOrigin(display_, gc_, c.X(), c.Y()); }
inline void X11::SetClipMask(Pixmap pixmap) const
	{ XSetClipMask(display_, gc_, pixmap); }
inline void X11::ClearArea(int x, int y, unsigned width, unsigned height,
                           Bool exposures) const
	{ XClearArea(display_, window_, x, y, width, height, exposures); }
inline void X11::ClearArea(Coord c, unsigned width, unsigned height,
                           Bool exposures) const
	{ ClearArea(c.X(), c.Y(), width, height, exposures); }
inline void
X11::CopyAreaToWindow(Drawable src, int src_x, int src_y, unsigned width,
                      unsigned height, int dest_x, int dest_y) const
{
	XCopyArea(display_, src, window_, gc_, src_x, src_y,
	          width, height, dest_x, dest_y);
}
inline void
X11::CopyAreaToWindow (Drawable src, int src_x, int src_y, unsigned width,
                       unsigned height,	Coord dest_c) const
{
	CopyAreaToWindow(src, src_x, src_y, width, height,
	                 dest_c.X(), dest_c.Y());
}
inline void X11::NextEvent(XEvent * evt) const
	{ XNextEvent(display_, evt); }
inline int X11::Pending() const
	{ return XPending(display_); }
inline void X11::Sync(Bool discard) const
	{ XSync(display_, discard); }
inline int
X11::CreatePixmapFromData(const char ** xpm, Pixmap * pixmap,
                          Pixmap * pixmask, XpmAttributes * attrs) const
{
	return XpmCreatePixmapFromData(display_, window_,
	                               const_cast<char**>(xpm),
	                               pixmap, pixmask, attrs);
}
inline void X11::FreePixmap(Pixmap pixmap) const
	{ XFreePixmap(display_, pixmap); }
inline XFontStruct * X11::LoadQueryFont(const std::string & font_name) const
	{ return XLoadQueryFont(display_, font_name.c_str()); }
inline void X11::FreeFont(XFontStruct * font) const
	{ XFreeFont(display_, font); }
inline void X11::SetFont(Font font) const
	{ XSetFont(display_, gc_, font); }
inline void X11::DrawString(Coord c, const std::string & str) const
{
	XDrawString(display_, window_, gc_, c.X(), c.Y(),
	            str.c_str(), str.size());
}


#endif
